import 'dart:ui' as ui;
import 'dart:ui';

import 'package:dartterm/assets.dart';
import 'package:dartterm/fonts.dart';
import 'package:dartterm/terminal.dart';
import 'package:flutter/material.dart';

class TerminalCustomPainter extends CustomPainter {
  Terminal t;
  double scalingFactor;

  static const int cellW = 8;
  static const int cellH = 8;

  TerminalCustomPainter(this.t, this.scalingFactor);

  @override
  bool shouldRepaint(TerminalCustomPainter oldDelegate) {
    return true;
  }

  @override
  void paint(Canvas canvas, Size size) {
    var pr = PictureRecorder();

    var smallSize = Size((cellW * Terminal.width).toDouble(),
        (cellH * Terminal.height).toDouble());

    var virtualCanvas = Canvas(pr, const Offset(0, 0) & smallSize);
    virtualPaint(virtualCanvas, smallSize);
    var image = pr
        .endRecording()
        // Renderer does something bizarre if I don't multiply by something greater than 1
        .toImageSync(cellW * Terminal.width * 2, cellH * Terminal.height * 2);
    canvas.drawImageRect(image, const Offset(0, 0) & smallSize,
        const Offset(0, 0) & size, Paint()..filterQuality = FilterQuality.none);
  }

  void virtualPaint(Canvas canvas, Size size) {
    final paint = Paint();

    var todos = Todos();

    // == Draw the background and foreground of every tile ==
    for (var i = 0; i < nTiles; i++) {
      final (pcxDst, pcyDst) = t.toXY(i) ?? (0, 0);
      final (pxDst, pyDst) = (pcxDst * cellW, pcyDst * cellH);
      final rectDst = Rect.fromLTWH(pxDst.toDouble(), pyDst.toDouble(),
          cellW.toDouble(), cellH.toDouble());

      final tile = t.tiles[i];

      var bgRect = Rect.fromLTWH(0, 0, cellW.toDouble(), cellH.toDouble());
      todos.add(Font.bg.imageName, bgRect, rectDst, tile.bg);

      var c = tile.content;
      if (c != null) {
        var fgRect = Rect.fromLTWH(c.sourceCx.toDouble() * cellW,
            c.sourceCy.toDouble() * cellH, cellW.toDouble(), cellH.toDouble());
        todos.add(c.sourceImage, fgRect, rectDst, tile.fg);
      }
    }

    for (var t in todos.imageTodos) {
      final atlas = t.atlas;
      if (atlas != null) {
        canvas.drawAtlas(atlas, t.transforms, t.rects, t.colors,
            BlendMode.modulate, null, paint);
      }
    }
  }
}

class Todos {
  List<ImageTodos> imageTodos = [];
  Map<String, int> imageTodoIx = {};

  void add(String source, Rect src, Rect dst, Color color) {
    var ix = imageTodoIx[source];
    if (ix == null) {
      ix = imageTodos.length;
      var image = assets.getImageIfAvailable(source);
      imageTodos.add(ImageTodos(image));
      imageTodoIx[source] = ix;
    }
    imageTodos[ix].add(src, dst, color);
  }
}

class ImageTodos {
  ui.Image? atlas;
  List<RSTransform> transforms = [];
  List<Rect> rects = [];
  List<Color> colors = [];

  ImageTodos(this.atlas);

  void add(Rect src, Rect dst, Color color) {
    transforms.add(RSTransform.fromComponents(
      rotation: 0.0,
      scale: 1.0,
      anchorX: 0.0,
      anchorY: 0.0,
      translateX: dst.left,
      translateY: dst.top,
    ));
    rects.add(src);
    colors.add(color);
  }
}
